"""
A module with internal/private models that are not thought to be used outside the MAUS package itself.
"""

from typing import Callable, List, Optional
from xml.etree.ElementTree import Element

import attrs
from lxml import etree  # type:ignore[import]

from maus.models.edifact_components import EdifactStack, EdifactStackQuery


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True)
class MigFilterResult:
    """
    the (internal) result of a query path search inside the tree
    """

    is_unique: Optional[bool]  #: True iff unique, None for no results, False for >1 result
    unique_result: Optional[Element]  #: unique element if there is any; None otherwise
    candidates: Optional[List[Element]]  #: list of candidates if there is >1 result


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True)
class EdifactStackSearchStrategy:
    """
    The search strategy allows to have a compact yet descriptive representation on how the edifact stack search works.
    The alternative to this is a very nested and hard to understand if/else/then structure with lots of branches.
    Any step inside the strategy has three possible outcomes which are represented by the :class:`_XQueryPathResult`:
    1. There is exactly one unique result => return/exit
    2. There are no results => start over again
    3. There are >1 results => apply additional filters
    """

    #: name, e.g. "filter by data element id"
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    #: the filter is the function that describes the strategy. It consumes the query and (optionally) a list of elements
    filter: Callable[[EdifactStackQuery, Optional[List[Element]]], MigFilterResult] = attrs.field(
        validator=attrs.validators.is_callable()
    )
    #: The unique result strategy is to return an edifact stack for the unique result element
    unique_result_strategy: Callable[[Element], EdifactStack] = attrs.field(validator=attrs.validators.is_callable())
    #: the no result strategy is to apply another filter based on those candidates that lead to no result (fallback)
    no_result_strategy: Optional["EdifactStackSearchStrategy"]
    #: in case of multiple results the next strategy uses the multiple results as input (sharpen)
    multiple_results_strategy: Optional["EdifactStackSearchStrategy"]

    def apply(self, query: EdifactStackQuery, pre_selection: Optional[List[Element]] = None) -> Optional[EdifactStack]:
        """
        Apply the defined strategy until we either have no ideas left or a unique result is found
        """
        # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
        # pylint: disable=not-callable
        filter_result: MigFilterResult = self.filter(query, pre_selection)
        if filter_result.is_unique is True:
            return self.unique_result_strategy(filter_result.unique_result)  # type:ignore[arg-type]
        if filter_result.candidates and len(filter_result.candidates) > 1:
            if self.multiple_results_strategy is not None:
                return self.multiple_results_strategy.apply(query, filter_result.candidates)
            return None
        if self.no_result_strategy is not None:
            return self.no_result_strategy.apply(query, pre_selection)
        return None



from datetime import datetime, timezone
from typing import Optional, Tuple

import pytest  # type:ignore[import]

from maus.edifact import (
    EdifactFormat,
    EdifactFormatVersion,
    get_edifact_format_version,
    is_edifact_boilerplate,
    pruefidentifikator_to_format,
)


class TestEdifact:
    """
    Tests the edifact module
    """

    @pytest.mark.parametrize(
        "expectation_tuple",
        [
            ("11042", EdifactFormat.UTILMD),
            ("13002", EdifactFormat.MSCONS),
            ("25001", EdifactFormat.UTILTS),
            ("10000", None),
        ],
    )
    def test_pruefi_to_format(self, expectation_tuple: Tuple[str, EdifactFormat]):
        """
        Tests that the prüfis can be mapped to an EDIFACT format
        """
        assert pruefidentifikator_to_format(expectation_tuple[0]) == expectation_tuple[1]

    @pytest.mark.parametrize(
        "key_date,expected_result",
        [
            pytest.param(
                datetime(2021, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                EdifactFormatVersion.FV2104,
                id="Anything before 2021-04-01",
            ),
            pytest.param(datetime(2021, 5, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2104),
            pytest.param(datetime(2021, 10, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2110),
            pytest.param(datetime(2022, 7, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2110),
            pytest.param(datetime(2022, 10, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2210),
            pytest.param(
                datetime(2050, 10, 1, 0, 0, 0, tzinfo=timezone.utc), EdifactFormatVersion.FV2304
            ),  # or what ever is the latest version
        ],
    )
    def test_format_version_from_keydate(self, key_date: datetime, expected_result: EdifactFormatVersion):
        actual = get_edifact_format_version(key_date)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "illegal_pruefi",
        [None, "", "asdas", "01234"],
    )
    def test_illegal_pruefis(self, illegal_pruefi: Optional[str]):
        """
        Test that illegal pruefis are not accepted
        :return:
        """
        with pytest.raises(ValueError):
            pruefidentifikator_to_format(illegal_pruefi)  # type:ignore[arg-type] # ok, because this raises an error

    @pytest.mark.parametrize(
        "segment_code,expected_is_boilerplate",
        [
            pytest.param(None, True),
            pytest.param("UNH", False),
            pytest.param("UNT", True),
            pytest.param("UNZ", True),
            pytest.param("DTM", False),
        ],
    )
    def test_is_boilerplate_segment(self, segment_code: Optional[str], expected_is_boilerplate: bool):
        actual_is_boilerplate = is_edifact_boilerplate(segment_code)
        assert actual_is_boilerplate == expected_is_boilerplate



from typing import Any, Optional

import pytest  # type:ignore[import]
from jsonpath_ng.ext import parse  # type:ignore[import] #  jsonpath is just installed in the tests

from maus import DeepAnwendungshandbuch
from maus.models.anwendungshandbuch import AhbMetaInformation
from maus.models.edifact_components import (
    DataElementFreeText,
    DataElementValuePool,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
)
from maus.transformation import generate_value_pool_replacement, transform_all_value_pools


def _get_value_using_json_path(data: dict, path: str) -> Optional[Any]:
    return parse(path).find(data)[0].value


def _set_value_using_json_path(data: dict, path: str, new_value: Any) -> None:
    current_datum = parse(path).find(data)[0]
    current_datum.full_path.update(data, new_value)


class DummyApplicationAccessor:
    # a class that obeys the type hints of the transformation.ApplicationAccessor
    def get_value(self, application_data: dict, application_discriminator: str) -> Optional[str]:
        """
        finds data in the non-edifact/application domain data model
        """
        return _get_value_using_json_path(application_data, application_discriminator)

    def set_value(self, _: dict, __: str) -> None:
        raise NotImplementedError("Not needed as of now")


class DummyEdifactAccessor:
    # a class that obeys the type hints of the transformation.EdifactAccessor
    def get_value(self, edifact_data: dict, edi_discriminator: str) -> Optional[str]:
        """
        finds data in the edifact data model ("EdiSeed")
        """
        return _get_value_using_json_path(edifact_data, edi_discriminator)

    def set_value(self, edifact_data: dict, edi_discriminator: str, new_value: Any) -> None:
        """
        finds data in the edifact domain data model ("EdiSeed") and replaces them with new_value.
        This modifies the edifact_data instance.
        """
        _set_value_using_json_path(data=edifact_data, path=edi_discriminator, new_value=new_value)


class DummyBo4eEdifactConverter:
    """this converter obeys the protocol of transformation.ApplicationEdifactConverter"""

    async def transform_application_to_edi_domain(self, application_domain_dict: dict):
        """
        Generate the edifact representations from the given application_domain_data.
        In real life this should call the transformation application➡edifact data model mapping logic.
        This is just a stub with hardcoded mappings.
        """
        # in real live this should call the transformer (edifact ↔ non-edifact)
        transaktionsgrund_application_domain = application_domain_dict["transaktionsdaten"]["transaktionsgrund"]
        application_to_edi_mapping_transaktionsgrund = {"EINZUG": "E01", "NEUANLAGE": "E02", "WECHSEL": "E03"}
        try:
            edifact_transaktionsgrund = application_to_edi_mapping_transaktionsgrund[
                transaktionsgrund_application_domain
            ]
        except KeyError:
            edifact_transaktionsgrund = None

        return {  # data in edifact domain
            "Transaktionsgrund": edifact_transaktionsgrund,
            "EdiFoo": "EdiBar",
            "TheThing": {"TheProperty": "Value1", "AnotherProperty": "AnotherValue"},
        }

    async def transform_edi_to_application_domain(self, edifact_seed: dict):
        """
        Generate the application representations from the given edifact seed.
        In real life this should call the transformation edifact➡application data model mapping logic.
        This is just a stub with hardcoded mappings.
        """
        transaktionsgrund_edifact_domain = edifact_seed["Transaktionsgrund"]
        edi_to_application_mapping_transaktionsgrund = {"E01": "EINZUG", "E02": "NEUANLAGE", "E03": "WECHSEL"}
        try:
            application_transaktionsgrund = edi_to_application_mapping_transaktionsgrund[
                transaktionsgrund_edifact_domain
            ]
        except KeyError:
            application_transaktionsgrund = None

        return {
            "noise": "foo",
            "transaktionsdaten": {
                "transaktionsgrund": application_transaktionsgrund,
            },
        }


class TestTransformation:
    async def test_value_pool_mapping_mwe(self):
        application_domain_data = {
            "noise": "foo",
            "transaktionsdaten": {
                "transaktionsgrund": None,
            },
        }  #: some data to start with, e.g. entered by a user in a frontend, hardcoded, whatever
        edi_to_application_data_model_mapping = {
            "$['Transaktionsgrund']": "$['transaktionsdaten']['transaktionsgrund']",
            "$['EdiFoo']": "$['stammdaten'][0]['applicationFoo']",
            "$['TheThing']['TheProperty']": "$['stammdaten'][0]['theProperty']",
            "$['TheThing']['AnotherProperty']": "$['stammdaten'][0]['anotherProperty']",
        }  #: a mapping that allows us to find data in both application and edifact domain

        data_element = DataElementValuePool(
            discriminator="$['Transaktionsgrund']",  # this is the original edifact domain discriminator
            data_element_id="1234",
            entered_input=None,
            value_pool=[
                ValuePoolEntry(meaning="the meaning of E01 from the AHB", ahb_expression="Muss [1]", qualifier="E01"),
                ValuePoolEntry(meaning="the meaning of E02 from the AHB", ahb_expression="Muss [2]", qualifier="E02"),
                ValuePoolEntry(meaning="the meaning of E03 from the AHB", ahb_expression="Muss [3]", qualifier="E03"),
            ],
        )
        converter = DummyBo4eEdifactConverter()
        edifact_accessor = DummyEdifactAccessor()
        application_accessor = DummyApplicationAccessor()
        value_mapping = await generate_value_pool_replacement(
            application_data_model=application_domain_data,
            data_element=data_element,
            converter=converter,
            edifact_accessor=edifact_accessor,
            application_accessor=application_accessor,
            edifact_to_non_edifact_path_mapping=edi_to_application_data_model_mapping,
        )
        data_element.replace_value_pool(value_mapping)
        assert data_element == DataElementValuePool(
            discriminator="$['Transaktionsgrund']",
            data_element_id="1234",
            entered_input=None,
            value_pool=[
                # here we got the application (e.g. bo4e) qualifiers with the respective expressions from the AHB
                ValuePoolEntry(
                    meaning="the meaning of E01 from the AHB", ahb_expression="Muss [1]", qualifier="EINZUG"
                ),
                ValuePoolEntry(
                    meaning="the meaning of E02 from the AHB", ahb_expression="Muss [2]", qualifier="NEUANLAGE"
                ),
                ValuePoolEntry(
                    meaning="the meaning of E03 from the AHB", ahb_expression="Muss [3]", qualifier="WECHSEL"
                ),
            ],
        )
        assert application_domain_data == {
            "noise": "foo",
            "transaktionsdaten": {
                "transaktionsgrund": None,
            },
        }  # the original data should not have been modified

    async def test_value_pool_mapping_in_deep_ahb(self):
        application_domain_data = {
            "noise": "foo",
            "transaktionsdaten": {
                "transaktionsgrund": None,
            },
        }  #: some data to start with, e.g. entered by a user in a frontend, hardcoded, whatever
        edi_to_application_data_model_mapping = {
            "$['Transaktionsgrund']": "$['transaktionsdaten']['transaktionsgrund']",
            "$['EdiFoo']": "$['stammdaten'][0]['applicationFoo']",
            "$['TheThing']['TheProperty']": "$['stammdaten'][0]['theProperty']",
            "$['TheThing']['AnotherProperty']": "$['stammdaten'][0]['anotherProperty']",
        }  #: a mapping that allows us to find data in both application and edifact domain

        deep_ahb = DeepAnwendungshandbuch(
            meta=AhbMetaInformation(pruefidentifikator="11042"),
            lines=[
                SegmentGroup(
                    ahb_expression="expr A",
                    discriminator="disc A",
                    segments=[
                        Segment(
                            ahb_expression="expr B",
                            discriminator="disc B",
                            data_elements=[
                                DataElementValuePool(
                                    discriminator="$['Transaktionsgrund']",
                                    # this is the original edifact domain discriminator
                                    data_element_id="1234",
                                    entered_input=None,
                                    value_pool=[
                                        ValuePoolEntry(
                                            meaning="the meaning of E01 from the AHB",
                                            ahb_expression="Muss [1]",
                                            qualifier="E01",
                                        ),
                                        ValuePoolEntry(
                                            meaning="the meaning of E02 from the AHB",
                                            ahb_expression="Muss [2]",
                                            qualifier="E02",
                                        ),
                                        ValuePoolEntry(
                                            meaning="the meaning of E03 from the AHB",
                                            ahb_expression="Muss [3]",
                                            qualifier="E03",
                                        ),
                                    ],
                                ),
                                DataElementFreeText(
                                    ahb_expression="Muss [1]",
                                    entered_input="Hello Mice",
                                    discriminator="bar",
                                    data_element_id="4567",
                                ),
                            ],
                        ),
                    ],
                    segment_groups=[],
                ),
            ],
        )
        converter = DummyBo4eEdifactConverter()
        edifact_accessor = DummyEdifactAccessor()
        application_accessor = DummyApplicationAccessor()
        await transform_all_value_pools(
            application_data_model=application_domain_data,
            deep_ahb=deep_ahb,
            converter=converter,
            edifact_accessor=edifact_accessor,
            application_accessor=application_accessor,
            edifact_to_non_edifact_path_mapping=edi_to_application_data_model_mapping,
        )
        assert deep_ahb.get_all_value_pools()[0] == DataElementValuePool(
            discriminator="$['Transaktionsgrund']",
            data_element_id="1234",
            entered_input=None,
            value_pool=[
                # here we got the application (e.g. bo4e) qualifiers with the respective expressions from the AHB
                ValuePoolEntry(
                    meaning="the meaning of E01 from the AHB", ahb_expression="Muss [1]", qualifier="EINZUG"
                ),
                ValuePoolEntry(
                    meaning="the meaning of E02 from the AHB", ahb_expression="Muss [2]", qualifier="NEUANLAGE"
                ),
                ValuePoolEntry(
                    meaning="the meaning of E03 from the AHB", ahb_expression="Muss [3]", qualifier="WECHSEL"
                ),
            ],
        )
        assert application_domain_data == {
            "noise": "foo",
            "transaktionsdaten": {
                "transaktionsgrund": None,
            },
        }  # the original data should not have been modified



from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestBeginnMsbMaus:
    """
    A unit test that ensures that the 11042/43/44 MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11042.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11042(self, datafiles):
        result = create_maus_and_assert(
            csv_path=Path(datafiles) / "11042.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11042_maus.json"),
        )
        kunde_des_msb_segments = result.maus.find_segments(
            group_predicate=lambda group: group.discriminator == "SG12",
            segment_predicate=lambda seg: seg.discriminator == "NAD"
            and seg.section_name == "Kunde des Messstellenbetreibers",
        )
        assert len(kunde_des_msb_segments) == 1
        nad_de_discriminators = [de.discriminator for de in kunde_des_msb_segments[0].data_elements]
        assert (
            '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Kunde des Messstellenbetreibers"][0]["Name des Beteiligten"]'
            in nad_de_discriminators
        )
        assert (
            '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Kunde des Messstellenbetreibers"][0]["Struktur"]'
            in nad_de_discriminators
        )
        korrespondenz_des_msb_kunden = result.maus.find_segments(
            group_predicate=lambda group: group.discriminator == "SG12",
            segment_predicate=lambda seg: seg.discriminator == "NAD"
            and seg.section_name == "Korrespondenzanschrift des Kunden des Messstellenbetreibers",
        )
        assert len(korrespondenz_des_msb_kunden) == 1
        nad_de_discriminators = [de.discriminator for de in korrespondenz_des_msb_kunden[0].data_elements]
        assert (
            '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Korrespondenzanschrift des Kunden des Messstellenbetreibers"][0]["Name des Beteiligten"]'
            in nad_de_discriminators
        )
        assert (
            '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Korrespondenzanschrift des Kunden des Messstellenbetreibers"][0]["Struktur"]'
            in nad_de_discriminators
        )
        assert (
            '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Korrespondenzanschrift des Kunden des Messstellenbetreibers"][0]["Strasse und Hausnummer oder Postfach"]'
            in nad_de_discriminators
        )
        assert (
            '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Korrespondenzanschrift des Kunden des Messstellenbetreibers"][0]["Ort"]'
            in nad_de_discriminators
        )
        assert (
            '$["Dokument"][0]["Nachricht"][0]["Vorgang"][0]["Korrespondenzanschrift des Kunden des Messstellenbetreibers"][0]["Postleitzahl"]'
            in nad_de_discriminators
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11043.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11043(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11043.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11043_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11044.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11044(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11044.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11044_maus.json"),
        )



from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestKuendigungMaus:
    """
    A unit test that ensures that the 11016/17/18 MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11016.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11016(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11016.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11016_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2210/UTILMD/11016.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11016_52e(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11016.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11016_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11017.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11017(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11017.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11017_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2210/UTILMD/11017.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11017_52e(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11017.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11017_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11018.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11018(self, datafiles):
        result = create_maus_and_assert(
            csv_path=Path(datafiles) / "11018.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11018_maus.json"),
        )
        assert (
            len(
                result.maus.find_segments(
                    segment_predicate=lambda seg: seg.section_name == "Datum des bereits bestaetigten Vertragsendes",
                )  # the "ae" instead of "ä" is somehow important, we don't wan't an 'ä' for reasons
            )
            > 0
        )  # https://github.com/Hochfrequenz/edifact-templates/pull/163 / ED4FTR-24952

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2210/UTILMD/11018.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11018_52e(self, datafiles):
        result = create_maus_and_assert(
            csv_path=Path(datafiles) / "11018.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11018_maus.json"),
        )
        assert (
            len(
                result.maus.find_segments(
                    segment_predicate=lambda seg: seg.section_name == "Datum des bereits bestaetigten Vertragsendes",
                )  # the "ae" instead of "ä" is somehow important, we don't wan't an 'ä' for reasons
            )
            > 0
        )  # https://github.com/Hochfrequenz/edifact-templates/pull/163 / ED4FTR-24952



"""
This module contains (static) functions that process single lxml.etree Elements.
Each function is separately unit tested.
"""
import re
from typing import List, Literal, Optional

# pylint:disable=no-name-in-module
from lxml.etree import Element  # type:ignore[import]

from maus.models._internal import MigFilterResult
from maus.models.edifact_components import EdifactStackQuery
from maus.reader.mig_ahb_name_helpers import are_similar_names


def get_segment_group_key_or_none(element: Element) -> Optional[str]:
    """
    returns the segment group of element if present; None otherwise
    """
    if "ref" in element.attrib and element.attrib["ref"].startswith("SG"):
        # the trivial case
        return element.attrib["ref"]
    return None


#: a regex to match a ref-segment: https://regex101.com/r/D81bbO/1
_single_nested_qualifier_pattern = re.compile(
    r"^(?P<segment_code>[A-Z]+):\d+:(?:\d+|\(\d+,\d+\))\[(?:\w+:)+\w+:?=(?P<qualifier>[A-Z\d]+)\]$"
)

#: a regex to match multiple ref/key segments: https://regex101.com/r/6XooRL/2
_multiple_nestes_qualifiers_pattern = re.compile(
    r"(?P<segment_code>[A-Z]+):\d+:(?:\d+|\(\d+,\d+\))\[(?P<inner>(?:(?:[A-Z]+:)?\d+:\d+=[A-Z\d]+\|?)+)\]$"
)


def get_nested_qualifiers(attrib_key: Literal["ref", "key"], element: Element) -> Optional[List[str]]:
    """
    returns the nested qualifier of an element if present; None otherwise
    we still return None instead of an empty list, because all the framework around this method actually check for None
    instead of empty lists
    """
    if attrib_key in element.attrib:
        body: str = element.attrib[attrib_key]
        single_match = _single_nested_qualifier_pattern.match(body)
        if single_match:
            return [single_match["qualifier"]]
        multi_match = _multiple_nestes_qualifiers_pattern.match(body)
        if multi_match:
            # if body == "QTY:1:1[1:0=265|1:0=Z10|1:0=Z08]"
            # then multi_match["inner"] is "1:0=265|1:0=Z10|1:0=Z08"
            return [expression.split("=")[1] for expression in multi_match["inner"].split("|")]
    return None


def get_ahb_name_or_none(element: Element) -> Optional[str]:
    """
    returns the ahbName of element if present; None otherwise
    """
    if "ahbName" in element.attrib:
        return element.attrib["ahbName"]
    return None


def filter_by_name(candidates: List[Element], query: EdifactStackQuery) -> List[Element]:
    """
    returns those elements that have the given name (in the query)
    """
    filtered_by_names = [
        x
        for x in candidates
        if are_similar_names(x.attrib["name"], query.name)
        or ("ahbName" in x.attrib and are_similar_names(x.attrib["ahbName"], query.name))
    ]
    return filtered_by_names


def filter_by_section_name(candidates: List[Element], query: EdifactStackQuery) -> List[Element]:
    """
    returns those elements that have the given section name (in the query)
    """
    filtered_by_names = [x for x in candidates if are_similar_names(x.attrib["name"], query.section_name)]
    return filtered_by_names


def list_to_mig_filter_result(candidates: List[Element]) -> MigFilterResult:
    """
    uses a given list of elements and differentiates three cases:
    1. the list contains a single unique element
    2. the list contains multiple elements
    3. the list contains no elements at all
    """
    if len(candidates) == 0:
        return MigFilterResult(candidates=None, is_unique=None, unique_result=None)
        # the == 1 case is handled last
    if len(candidates) > 1:
        return MigFilterResult(candidates=candidates, is_unique=False, unique_result=None)
    return MigFilterResult(candidates=None, is_unique=True, unique_result=candidates[0])



from typing import List, Literal, Optional

import pytest  # type:ignore[import]
from lxml import etree  # type:ignore[import]

from maus.models.edifact_components import EdifactStackQuery
from maus.reader.etree_element_helpers import (
    filter_by_name,
    filter_by_section_name,
    get_ahb_name_or_none,
    get_nested_qualifiers,
    get_segment_group_key_or_none,
    list_to_mig_filter_result,
)


def _string_to_element(xml_elem: str) -> etree.Element:
    """
    helper method that adds an XML preamble to the given string, parses the concatenated string as etree
    and returns the root element
    """
    valid_xml = '<?xml version="1.0"?>' + xml_elem
    root = etree.fromstring(valid_xml)
    return root


def _string_to_elements(xml_elem: str) -> List[etree.Element]:
    """
    helper method that adds an XML preamble and virtual root to the given string,
    parses the concatenated string as etree and returns the elements on root level
    """
    valid_xml = '<?xml version="1.0"?><virtual_root>' + xml_elem + "</virtual_root>"
    root = etree.fromstring(valid_xml)
    return [x for x in root.findall("./")]


class TestEtreeSingleElementHelpers:
    """
    Tests the behaviour of static helper methods (for single elements)
    """

    @pytest.mark.parametrize(
        "xml_string, expected",
        [
            pytest.param('<foo ref="SG14"></foo>', "SG14"),
            pytest.param("<bar></bar>", None),
        ],
    )
    def test_get_segment_group_key_or_none(self, xml_string: str, expected: Optional[str]):
        element = _string_to_element(xml_string)
        assert get_segment_group_key_or_none(element) == expected

    @pytest.mark.parametrize(
        "xml_string, expected",
        [
            pytest.param('<foo ahbName="X"></foo>', "X"),
            pytest.param("<bar></bar>", None),
        ],
    )
    def test_get_ahb_name_or_none(self, xml_string: str, expected: Optional[str]):
        element = _string_to_element(xml_string)
        assert get_ahb_name_or_none(element) == expected

    @pytest.mark.parametrize(
        "ref_or_key, xml_string, expected",
        [
            pytest.param("ref", '<foo ref="hello"></foo>', None),
            pytest.param("ref", '<bar ref="NAD:2:0[1:0=MS]"></bar>', ["MS"]),
            pytest.param("ref", '<baz ref="RFF:1:1[RFF:1:0=Z13]"></baz>', ["Z13"]),
            pytest.param("ref", '<baz ref="DTM:1:1[1:0=469]"></baz>', ["469"]),
            pytest.param("key", '<baz asd="hello"></baz>', None),
            pytest.param("key", '<foo key="NAD:2:0[1:0=MS]"></foo>', ["MS"]),
            pytest.param("key", '<foo key="QTY:1:1[1:0=265|1:0=Z10|1:0=Z08]"></foo>', ["265", "Z10", "Z08"]),
            pytest.param("ref", '<field ref="FTX:4:(0,4)[1:0=ACB]" />', ["ACB"]),
            pytest.param("ref", '<field ref="CAV:1:0[CCI:1:0=Z02|CCI:1:0=Z04]" />', ["Z02", "Z04"]),
            pytest.param("key", '<class key="NAD:5:(0,3)[NAD:1:0=Z08]" ref="SG12" />', ["Z08"]),
            pytest.param("key", '<class key="NAD:4:(0,4)[NAD:1:0=Z07]" ref="SG12" />', ["Z07"]),
        ],
    )
    def test_get_nested_qualifier(self, ref_or_key: Literal["ref", "key"], xml_string: str, expected: List[str]):
        element = _string_to_element(xml_string)
        assert get_nested_qualifiers(ref_or_key, element) == expected


class TestMultipleElementsHelpers:
    """
    Tests the behaviour of static helper methods (for multiple elements)
    """

    @pytest.mark.parametrize(
        "xml_string,expected_is_unique, expected_has_unique_result, expected_candidates_length",
        [
            pytest.param(
                "<bar/><baz/><foo/>",
                False,
                None,
                3,
            ),
            pytest.param(
                "<bar/>",
                True,
                True,
                None,
            ),
            pytest.param(
                "",
                None,
                False,
                None,
            ),
        ],
    )
    def test_list_to_mig_filter_result(
        self,
        xml_string: str,
        expected_is_unique: Optional[bool],
        expected_has_unique_result: bool,
        expected_candidates_length: Optional[int],
    ):
        elements = _string_to_elements(xml_string)
        actual = list_to_mig_filter_result(elements)
        assert actual.is_unique is expected_is_unique
        if expected_is_unique:
            assert actual.unique_result is not None
        else:
            assert actual.unique_result is None
        if expected_candidates_length is not None:
            assert actual.candidates is not None
            assert len(actual.candidates) == expected_candidates_length
        else:
            assert actual.candidates is None

    @pytest.mark.parametrize(
        "xml_string, query, expected_allowed_ids",
        [
            pytest.param(
                '<bar name="foo" id="this"></bar><baz name="fuu" id="that"/>',
                EdifactStackQuery(
                    segment_group_key="SG1",
                    segment_code="ASD",
                    data_element_id="1234",
                    name="foo",
                    predecessor_qualifier=None,
                ),
                ["this"],
            ),
            pytest.param(
                '<bar name="bar" id="this"></bar><baz name=" Foo-" id="that"/>',
                EdifactStackQuery(
                    segment_group_key="SG1",
                    segment_code="ASD",
                    data_element_id="1234",
                    name="foo",
                    predecessor_qualifier=None,
                ),
                ["that"],
            ),
            pytest.param(
                '<bar name="bar" id="this"></bar><baz name="something" ahbName=" Foo-" id="that"/>',
                EdifactStackQuery(
                    segment_group_key="SG1",
                    segment_code="ASD",
                    data_element_id="1234",
                    name="foo",
                    predecessor_qualifier=None,
                ),
                ["that"],
            ),
        ],
    )
    def test_filter_by_name(self, xml_string: str, query: EdifactStackQuery, expected_allowed_ids: List[str]):
        elements = _string_to_elements(xml_string)
        actual = filter_by_name(elements, query)
        assert [x.attrib["id"] for x in actual] == expected_allowed_ids

    @pytest.mark.parametrize(
        "xml_string, query, expected_allowed_ids",
        [
            pytest.param(
                '<bar name="foo" id="this"></bar><baz name="fuu" id="that"/>',
                EdifactStackQuery(
                    segment_group_key="SG12",
                    section_name="Foo",
                    segment_code="ASD",
                    data_element_id="1234",
                    name="asd",
                    predecessor_qualifier=None,
                ),
                ["this"],
            ),
            pytest.param(
                '<bar name="hello" id="this"></bar><baz name=" Foo-" id="that"/>',
                EdifactStackQuery(
                    segment_group_key="Foo",
                    section_name="Hello",
                    segment_code="ASD",
                    data_element_id="1234",
                    name="asd",
                    predecessor_qualifier=None,
                ),
                ["this"],
            ),
            pytest.param(
                '<bar name="hello" id="this"></bar><baz name="HeLlO" id="that"/>',
                EdifactStackQuery(
                    segment_group_key="Foo",
                    section_name="Hello",
                    segment_code="ASD",
                    data_element_id="1234",
                    name="asd",
                    predecessor_qualifier=None,
                ),
                ["this", "that"],
            ),
        ],
    )
    def test_filter_by_section_name(self, xml_string: str, query: EdifactStackQuery, expected_allowed_ids: List[str]):
        elements = _string_to_elements(xml_string)
        actual = filter_by_section_name(elements, query)
        assert [x.attrib["id"] for x in actual] == expected_allowed_ids



import json
from pathlib import Path
from sys import gettrace

import attrs

from maus import DeepAnwendungshandbuch, SegmentGroupHierarchy, to_deep_ahb
from maus.deep_ahb_mig_joiner import replace_discriminators_with_edifact_stack
from maus.models.anwendungshandbuch import DeepAnwendungshandbuchSchema, FlatAnwendungshandbuch
from maus.models.message_implementation_guide import SegmentGroupHierarchySchema
from maus.reader.flat_ahb_reader import FlatAhbCsvReader
from maus.reader.mig_reader import MigXmlReader


def is_in_debug_mode() -> bool:
    """
    returns true if it is called in debug mode.
    Generally it's wrong to behave differently depending on "from where" the code is called.
    In our test cases we want to know if the tests run in debug mode in order to decide if testdata should be actively
    updated.
    :return: true iff debug mode
    """
    return gettrace() is not None


def should_write_to_submodule() -> bool:
    """
    returns true if tests should write to the submodule (e.g. updated MAUS s' or AHBs)
    """
    # return True # enable this line to overwrite the test data
    return is_in_debug_mode()


def write_to_file_or_assert_equality(deep_ahb: DeepAnwendungshandbuch, file_path: Path) -> None:
    """
    writes the given maus/deep ahb into the provided file path or read it from there and assert equality.
    The decision (write vs. read) is met by 'should_write_to_submodule'
    :return:
    """
    write_into_submodule: bool = should_write_to_submodule()
    schema = DeepAnwendungshandbuchSchema()
    if write_into_submodule:
        with open(file_path, "w+", encoding="utf-8") as outfile:
            json_dict = schema.dump(deep_ahb)
            json.dump(json_dict, outfile, ensure_ascii=True, indent=2, sort_keys=True)
    else:
        with open(file_path, "r", encoding="utf-8") as infile:
            json_content = json.load(infile)
            expected_maus = schema.load(json_content)
        assert deep_ahb == expected_maus


@attrs.define(kw_only=True)
class IntegrationTestResult:
    """
    A class that contains test results
    """

    flat_ahb: FlatAnwendungshandbuch = attrs.field(validator=attrs.validators.instance_of(FlatAnwendungshandbuch))
    segment_group_hierarchy: SegmentGroupHierarchy = attrs.field(
        validator=attrs.validators.instance_of(SegmentGroupHierarchy)
    )
    deep_ahb: DeepAnwendungshandbuch = attrs.field(validator=attrs.validators.instance_of(DeepAnwendungshandbuch))
    maus: DeepAnwendungshandbuch = attrs.field(validator=attrs.validators.instance_of(DeepAnwendungshandbuch))


def create_maus_and_assert(
    csv_path: Path, sgh_path: Path, template_path: Path, maus_path: Path
) -> IntegrationTestResult:
    """
    The repetitive part of every integration test so far:
    read the CSV into a FlatAhb, read the SGH from the file, read the template, join both AHB and MIG into a MAUS
    """
    reader = FlatAhbCsvReader(file_path=csv_path)
    flat_ahb = reader.to_flat_ahb()
    with open(sgh_path, "r", encoding="utf-8") as sgh_file:
        sgh = SegmentGroupHierarchySchema().loads(sgh_file.read())
    actual_deep_ahb = to_deep_ahb(flat_ahb, sgh)
    actual_json = DeepAnwendungshandbuchSchema().dumps(actual_deep_ahb, ensure_ascii=True, sort_keys=True)
    mig_reader = MigXmlReader(template_path)
    maus = DeepAnwendungshandbuchSchema().loads(actual_json)  # maus is a copy of the (unchanged) actual_deep_ahb
    assert mig_reader is not None
    replace_discriminators_with_edifact_stack(maus, mig_reader, ignore_errors=False)
    actual_maus_json = DeepAnwendungshandbuchSchema().dumps(maus, ensure_ascii=True, sort_keys=True, indent=True)
    assert actual_maus_json is not None
    write_to_file_or_assert_equality(maus, maus_path)
    result = IntegrationTestResult(flat_ahb=flat_ahb, segment_group_hierarchy=sgh, deep_ahb=actual_deep_ahb, maus=maus)
    return result



from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestKuendigungMsbMaus:
    """
    A unit test that ensures that the 11039/40/41 MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11039.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11039(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11039.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11039_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11040.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11040(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11040.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11040_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11041.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11041(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11041.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11041_maus.json"),
        )



from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestEogMaus:
    """
    A unit test that ensures that the 11013/14/16 MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11013.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11013(self, datafiles):
        result = create_maus_and_assert(
            csv_path=Path(datafiles) / "11013.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11013_maus.json"),
        )
        assert (
            len(
                result.maus.find_segments(
                    segment_predicate=lambda seg: seg.discriminator == "SEQ"
                    and seg.section_name == "OBIS-Daten der Marktlokation",
                )
            )
            > 0
        )  # https://github.com/Hochfrequenz/edifact-templates/issues/81

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2210/UTILMD/11013.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11013_52e(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11013.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11013_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11014.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11014(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11014.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11014_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2210/UTILMD/11014.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11014_52e(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11014.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11014_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11015.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11015(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11015.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11015_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2e.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2210/UTILMD/11015.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2210/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11015_52e(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11015.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2e.template"),
            maus_path=Path("edifact-templates/maus/FV2210/UTILMD/11015_maus.json"),
        )



from typing import List, Optional, Tuple

import pytest  # type:ignore[import]
from lxml import etree  # type:ignore[import]
from lxml.etree import Element  # type:ignore[import]

from maus.models._internal import MigFilterResult
from maus.models.edifact_components import EdifactStack, EdifactStackLevel, EdifactStackQuery
from maus.reader.etree_element_helpers import list_to_mig_filter_result
from maus.reader.mig_reader import MigXmlReader


class TestMigXmlReaderMwe:
    """
    Tests the behaviour of the Message Implementation Guide model with manually crafted minimal working examples.
    """

    # todo: you could probably use an indirect parametrization instead of the static helper methods below

    @staticmethod
    def _prepare_xml_reader_and_element(xml_string: str, element_xpath: str) -> Tuple[MigXmlReader, Element]:
        """
        A helper method for easier test setup
        """
        reader = MigXmlReader(xml_string)
        elements = reader._original_tree.xpath(element_xpath)
        if len(elements) != 1:
            raise ValueError(
                f"In this test you're thought to provide an xpath that matches exactly one element of the given XML but found {len(elements)}."
            )
        element = elements[0]
        return reader, element

    @staticmethod
    def _prepare_xml_reader_and_elements(
        xml_string: str, elements_xpathes: List[str]
    ) -> Tuple[MigXmlReader, List[Element]]:
        """
        A helper method for easier test setup
        """
        reader = MigXmlReader(xml_string)
        result_elements: List[Element] = []
        for xpath in elements_xpathes:
            elements = reader._original_tree.xpath(xpath)
            if len(elements) != 1:
                raise ValueError(
                    f"In this test you're thought to provide an xpath that matches exactly one element of the given XML but found {len(elements)} for xpath '{xpath}'."
                    # this means, the paths should be absolute
                )
            element = elements[0]
            result_elements.append(element)
        return reader, result_elements

    @staticmethod
    def _prepare_mig_filter_result(tree: etree.ElementTree, elements_xpathes: List[str]) -> MigFilterResult:
        """
        A helper method for easier test setup. It basically re-implements "get_unique_result_by_xpath"
        """
        results: List[Element] = []
        for xpath in elements_xpathes:
            elements = tree.xpath(xpath)
            if len(elements) != 1:
                raise ValueError(
                    f"In this test you're thought to provide an xpath that matches exactly one element of the given XML but found {len(elements)} for xpath '{xpath}'."
                )
            element = elements[0]
            results.append(element)
        return list_to_mig_filter_result(results)

    @pytest.mark.parametrize(
        "xml_string, expected_format",
        [
            pytest.param('<?xml version="1.0"?><MSCONS></MSCONS>', "MSCONS"),
            pytest.param('<?xml version="1.0"?><UTILMD></UTILMD>', "UTILMD"),
        ],
    )
    def test_get_edifact_format(self, xml_string: str, expected_format: str):
        reader = MigXmlReader(xml_string)
        assert reader.get_format_name() == expected_format

    @pytest.mark.parametrize(
        "xml_string, element_xpath, use_sanitized, expected_result",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42"><foo/></class></MSCONS>',
                "//MSCONS/class/foo",
                False,
                "SG42",
            ),
        ],
    )
    def test_get_parent_segment_group(
        self, xml_string: str, element_xpath: str, use_sanitized: bool, expected_result: Optional[str]
    ):
        reader, element = TestMigXmlReaderMwe._prepare_xml_reader_and_element(xml_string, element_xpath)
        actual = reader.get_parent_segment_group_key(element, use_sanitized_tree=use_sanitized)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string, element_xpath, use_sanitized, expected_result",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><foo key="DTM:1:2[1:0=Z21]"><bar/></foo></MSCONS>',
                "//MSCONS/foo/bar",
                False,
                ["Z21"],
                id="single predecessor",
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class key="QTY:1:1[1:0=265|1:0=Z10|1:0=Z08]"><field /></class></UTILMD>',
                "//UTILMD/class/field",
                False,
                ["265", "Z10", "Z08"],
                id="multiple predecessors",
            ),
        ],
    )
    def test_get_parent_predecessors(
        self, xml_string: str, element_xpath: str, use_sanitized: bool, expected_result: List[str]
    ):
        reader, element = TestMigXmlReaderMwe._prepare_xml_reader_and_element(xml_string, element_xpath)
        actual = reader.get_parent_predecessors(element, use_sanitized_tree=use_sanitized)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string, element_xpath, use_sanitized, expected_result",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><foo name="asd"><bar ahbName="qwe"/></foo></MSCONS>',
                "//MSCONS/foo/bar",
                False,
                EdifactStack.from_json_path(r'$["asd"]["qwe"]'),
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class name="asd"><bar ahbName="qwe"/></class></MSCONS>',
                "//MSCONS/class/bar",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="asd", is_groupable=True),
                        EdifactStackLevel(name="qwe", is_groupable=False),
                    ]
                ),
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ahbName="asd"><bar name="qwe"/><class ahbName="opi"/></class></MSCONS>',
                "//MSCONS/class/class",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="asd", is_groupable=True),
                        EdifactStackLevel(name="opi", is_groupable=True),
                    ]
                ),
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class name="Dokument" ref="/" key="UNB:5:0"><class name="Absender_NAD" migName="MP-ID Absender" max="1" ref="SG2" key="NAD:2:0[1:0=MS]" ahbName="MP-ID Absender" meta.type="group" meta.objType="Marktpartner"></class></class></UTILMD>',
                "//UTILMD/class/class",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="Dokument", is_groupable=True),
                        EdifactStackLevel(name="MP-ID Absender", is_groupable=True),
                        # "MP-ID Absender" shall be preferred over "Absender_NAD"
                    ]
                ),
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class name="Dokument" ref="/" key="UNB:5:0"><class name="Nachricht" ref="UNH" key="UNH:1:0" max="9999" meta.type="group"><field name="Nachrichten-Referenznummer" ref="UNH:1:0" meta.id="0062" /><field name="Kategorie" migName="Beginn der Nachricht" meta.id="1001" ref="BGM:1:0" groupBy="Beginn der Nachricht" meta.help="Hier wird die Kategorie der gesamten Nachricht für alle Vorgänge angegeben." meta.type="repository" meta.typeInfo="Kategorierepository"/></class></class></UTILMD>',
                "//UTILMD/class/class/field[2]",
                False,
                EdifactStack(
                    levels=[
                        EdifactStackLevel(name="Dokument", is_groupable=True),
                        EdifactStackLevel(name="Nachricht", is_groupable=True),
                        EdifactStackLevel(name="Kategorie", is_groupable=False),
                    ]
                ),
            ),
        ],
    )
    def test_element_to_edifact_stack(
        self, xml_string: str, element_xpath: str, use_sanitized: bool, expected_result: EdifactStack
    ):
        reader, element = TestMigXmlReaderMwe._prepare_xml_reader_and_element(xml_string, element_xpath)
        actual = reader.element_to_edifact_stack(element, use_sanitized_tree=use_sanitized)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42"><foo/></class></MSCONS>',
                ["//MSCONS/class/foo"],
                EdifactStackQuery(
                    segment_group_key="SG42",
                    segment_code="XXX",
                    data_element_id="0000",
                    name="",
                ),
                ["//MSCONS/class/foo"],
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42"><foo/></class></MSCONS>',
                ["//MSCONS/class/foo"],
                EdifactStackQuery(
                    segment_group_key="SG17",
                    segment_code="XXX",
                    data_element_id="0000",
                    name="",
                ),
                [],
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42"><foo/><bar/></class></MSCONS>',
                ["//MSCONS/class/foo", "//MSCONS/class/bar"],
                EdifactStackQuery(
                    segment_group_key="SG42",
                    segment_code="XXX",
                    data_element_id="0000",
                    name="",
                ),
                ["//MSCONS/class/foo", "//MSCONS/class/bar"],
            ),
        ],
    )
    def test_get_unique_result_by_segment_group(
        self, xml_string: str, candidates_xpaths: List[str], query: EdifactStackQuery, expected_result_xpaths: List[str]
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_unique_result_by_segment_group(candidates=candidates, query=query, use_sanitized_tree=False)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42"><foo/></class><class ref="SG77"><bar/></class></MSCONS>',
                ["//MSCONS/class/foo", "//MSCONS/class/bar"],
                EdifactStackQuery(
                    segment_group_key="SG42",
                    segment_code="XXX",  # dummy
                    data_element_id="0000",  # dummy
                    name="",
                ),
                ["//MSCONS/class/foo"],
            ),
        ],
    )
    def test_get_unique_result_by_parent_segment_group(
        self, xml_string: str, candidates_xpaths: List[str], query: EdifactStackQuery, expected_result_xpaths: List[str]
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_unique_result_by_parent_segment_group(candidates=candidates, query=query)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class name="Kunde des Messstellenbetreibers" ref="SG12" key="NAD:4:(0,4)[NAD:1:0=Z07]" groupKey="Name des Beteiligten" meta.objType="Personen"><field name="Name des Beteiligten" ref="NAD:4:0" meta.id="3036" meta.orderId="2" /><field name="Name des Beteiligten1" ref="NAD:4:1" meta.id="3036" meta.orderId="3" /><field name="Name des Beteiligten2" ref="NAD:4:2" meta.id="3036" meta.orderId="4" /><field name="Name des Beteiligten3" ref="NAD:4:3" meta.id="3036" meta.orderId="5" /><field name="Name des Beteiligten4" ref="NAD:4:4" meta.id="3036" meta.orderId="6" /></class></UTILMD>',
                [
                    "//UTILMD/class/field[1]",
                    "//UTILMD/class/field[2]",
                    "//UTILMD/class/field[3]",
                    "//UTILMD/class/field[4]",
                    "//UTILMD/class/field[5]",
                ],
                EdifactStackQuery(
                    segment_group_key="SG12",
                    segment_code="NAD",
                    data_element_id="3036",
                    name="Name",
                    section_name="Kunde des Messstellenbetreibers",
                    predecessor_qualifier="Z07",
                ),
                ["//UTILMD/class/field[1]"],
                id="Kunde des Messstellenbetreibers/Name",
            ),
        ],
    )
    def test_get_first_if_candidates_only_differ_by_trailing_number(
        self, xml_string: str, candidates_xpaths: List[str], query: EdifactStackQuery, expected_result_xpaths: List[str]
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_first_if_candidates_only_differ_by_trailing_number(candidates=candidates, query=query)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><REQOTE><class ahbName="MP-ID Absender"><foo/></class><class ahbName="MP-ID Empfänger"><bar/></class></REQOTE>',
                ["//REQOTE/class/foo", "//REQOTE/class/bar"],
                EdifactStackQuery(
                    segment_group_key="SG42",  # dummy
                    segment_code="XXX",  # dummy
                    data_element_id="0000",  # dummy
                    name="",
                    section_name="MP-ID Absender",
                ),
                ["//REQOTE/class/foo"],
            ),
        ],
    )
    def test_get_unique_result_by_parent_ahb_name_section_name(
        self, xml_string: str, candidates_xpaths: List[str], query: EdifactStackQuery, expected_result_xpaths: List[str]
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_unique_result_by_parent_ahb_name_section_name(candidates=candidates, query=query)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42"><foo name="hello"/><bar name="world"/></class></MSCONS>',
                ["//MSCONS/class/foo", "//MSCONS/class/bar"],
                EdifactStackQuery(
                    segment_group_key="SG123",  # dummy
                    segment_code="XXX",  # dummy
                    data_element_id="0000",  # dummy
                    name="world",
                ),
                ["//MSCONS/class/bar"],
            ),
        ],
    )
    def test_get_unique_result_by_name(
        self, xml_string: str, candidates_xpaths: List[str], query: EdifactStackQuery, expected_result_xpaths: List[str]
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_unique_result_by_name(candidates=candidates, query=query)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42" name="foo"><x name="xx"/><y name="yy"/></class></MSCONS>',
                ["//MSCONS/class/x", "//MSCONS/class/y", "//MSCONS/class/y"],
                EdifactStackQuery(
                    segment_group_key="SG123",  # dummy
                    segment_code="XXX",  # dummy
                    data_element_id="0000",  # dummy
                    name="asd",  # dummy
                    section_name="xx",
                ),
                ["//MSCONS/class/x"],
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42" name="foo"><x name="XxX"/><y name="yy"/></class></MSCONS>',
                ["//MSCONS/class/x", "//MSCONS/class/y", "//MSCONS/class/y"],
                EdifactStackQuery(
                    segment_group_key="SG123",  # dummy
                    segment_code="XXX",  # dummy
                    data_element_id="0000",  # dummy
                    name="asd",  # dummy
                    section_name="xxx",
                ),
                ["//MSCONS/class/x"],
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG42" name="foo"><x name=" XxX-"/><y name="yy"/></class></MSCONS>',
                ["//MSCONS/class/x", "//MSCONS/class/y", "//MSCONS/class/y"],
                EdifactStackQuery(
                    segment_group_key="SG123",  # dummy
                    segment_code="XXX",  # dummy
                    data_element_id="0000",  # dummy
                    name="asd",  # dummy
                    section_name="xxx",
                ),
                ["//MSCONS/class/x"],
            ),
        ],
    )
    def test_get_unique_result_by_section_name(
        self, xml_string: str, candidates_xpaths: List[str], query: EdifactStackQuery, expected_result_xpaths: List[str]
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_unique_result_by_section_name(candidates=candidates, query=query)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class ref="SG1" key="RFF:1:1[RFF:1:0=AGI]"><foo/></class></MSCONS>',
                ["//MSCONS/class"],
                EdifactStackQuery(
                    segment_group_key="SG1",
                    segment_code="RFF",
                    predecessor_qualifier="AGI",
                    data_element_id="0000",  # dummy
                    name="",  # dummy
                ),
                ["//MSCONS/class"],
                id="rffuck1 mit ref",
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class><foo key="RFF:1:1[RFF:1:0=AGI]"/></class></MSCONS>',
                ["//MSCONS/class/foo"],
                EdifactStackQuery(
                    segment_group_key="SG1",
                    segment_code="RFF",
                    predecessor_qualifier="AGI",
                    data_element_id="0000",  # dummy
                    name="",  # dummy
                ),
                ["//MSCONS/class/foo"],
                id="rffuck2 mit key",
            ),
            pytest.param(
                '<?xml version="1.0"?><MSCONS><class><foo ref="DTM:1:1[1:0=Z21]"/></class></MSCONS>',
                ["//MSCONS/class/foo"],
                EdifactStackQuery(
                    segment_group_key="SG1",
                    segment_code="DTM",
                    predecessor_qualifier="Z21",
                    data_element_id="0000",  # dummy
                    name="",  # dummy
                ),
                ["//MSCONS/class/foo"],
                id="dtm",
            ),
        ],
    )
    def test_get_unique_result_by_predecessor(
        self, xml_string: str, candidates_xpaths: List[str], query: EdifactStackQuery, expected_result_xpaths: List[str]
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_unique_result_by_predecessor(candidates=candidates, query=query)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "xml_string,candidates_xpaths,query,skip_levels,expected_result_xpaths",
        [
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class key="LOC:2:0[1:0=Z08]"><field ref="LOC:2:0"/></class></UTILMD>',
                ["//UTILMD/class/field"],
                EdifactStackQuery(
                    segment_group_key="SG1",  # dummy
                    segment_code="LOC",
                    predecessor_qualifier="Z08",
                    data_element_id="0000",  # dummy
                    name="",  # dummy
                ),
                0,
                ["//UTILMD/class/field"],
                id="loc",
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><asd key="FOO:4:5[1:0=Z11]"><class key="LOC:2:0[1:0=Z08]"><field ref="LOC:2:0"/></class></asd><yxc key="FOO:4:5[1:0=Z22]"><class key="LOC:2:0[1:0=Z08]"><field ref="LOC:2:0"/></class></yxc></UTILMD>',
                ["//UTILMD/asd/class/field", "//UTILMD/yxc/class/field"],
                EdifactStackQuery(
                    segment_group_key="SG1",  # dummy
                    segment_code="LOC",
                    predecessor_qualifier="Z11",
                    data_element_id="0000",  # dummy
                    name="",  # dummy
                ),
                1,
                ["//UTILMD/asd/class/field"],
                id="loc",
            ),
            pytest.param(
                '<?xml version="1.0"?><UTILMD><class name="Korrespondenzanschrift des Kunden des Lieferanten" ref="SG12" key="NAD:5:(0,3)[NAD:1:0=Z04]" meta.objType="Personen"><field name="Strasse und Hausnummer oder Postfach" ref="NAD:5:0" meta.id="3042" /></class><class name="Korrespondenzanschrift des Kunden des Messstellenbetreibers" ahbName="Korrespondenzanschrift des Kunden des MSB" ref="SG12" key="NAD:5:(0,3)[NAD:1:0=Z08]" meta.objType="Personen"><field name="Strasse und Hausnummer oder Postfach" ref="NAD:5:0" meta.id="3042" /></class></UTILMD>',
                ["//UTILMD/class[1]/field", "//UTILMD/class[2]/field"],
                EdifactStackQuery(
                    segment_group_key="SG12",  # dummy
                    segment_code="NAD",
                    predecessor_qualifier="Z08",
                    data_element_id="3042",  # dummy
                    name="Straße und Hausnummer oderPostfach",
                    section_name="Korrespondenzanschrift des Kunden des Messstellenbetreibers",
                ),
                0,
                ["//UTILMD/class[2]/field"],
                id="anschrift des msb-kunden",
            ),
        ],
    )
    def test_get_unique_result_by_parent_predecessor(
        self,
        xml_string: str,
        candidates_xpaths: List[str],
        query: EdifactStackQuery,
        skip_levels: int,
        expected_result_xpaths: List[str],
    ):
        reader, candidates = TestMigXmlReaderMwe._prepare_xml_reader_and_elements(xml_string, candidates_xpaths)
        expected_result = TestMigXmlReaderMwe._prepare_mig_filter_result(reader._original_tree, expected_result_xpaths)
        actual = reader.get_unique_result_by_parent_predecessor(
            candidates=candidates, query=query, skip_levels=skip_levels
        )
        assert actual == expected_result
